<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="zh">

<head>
  <title>{{ i18n.cover }}</title>
</head>

<body>
  <p>
    <img src="../assets/cover.png" alt="Cover" />
  </p>
</body>

</html>